// -*- C++ -*-
#include "Rivet/Analysis.hh"
#include "Rivet/Projections/FastJets.hh"
#include "Rivet/Tools/Logging.hh"
#include "Rivet/Projections/FinalState.hh"
#include "Rivet/Projections/IdentifiedFinalState.hh"
#include "Rivet/Projections/WFinder.hh"
#include "Rivet/Projections/LeadingParticlesFinalState.hh"
#include "Rivet/Projections/UnstableFinalState.hh"
#include "Rivet/Projections/VetoedFinalState.hh"
#include "Rivet/Tools/ParticleIdUtils.hh"
#include "Rivet/Particle.hh"
#include "Rivet/Projections/DressedLeptons.hh"
#include "Rivet/Projections/MergedFinalState.hh"
#include "Rivet/Projections/MissingMomentum.hh"
#include "Rivet/Projections/InvMassFinalState.hh"
#include "HepMC/GenEvent.h"

namespace Rivet {


   double Zmass = 91.1876;  // GeV
   const double pi = acos(-1);


   /// generic Z candidate
   class Zstate : public ParticlePair  {

      public: 
         Zstate(){}

         Zstate(ParticlePair _particlepair):
            ParticlePair(_particlepair) {}

         FourMomentum momentum() const {
            return first.momentum() + second.momentum();
         }

   };

   struct Zstate_less_pT
   {
      bool operator()( const Zstate& lx, const Zstate& rx ) const {
         return ( lx.momentum().pT() > rx.momentum().pT());
      }
   };


   /// 4l to ZZ assignment -- algorithm
   void identifyZstates( Zstate & Z1, Zstate &Z2, const Particles & leptons_sel4l) {

      /////////////////////////////////////////////////////////////////////////////
      /// ZZ->4l pairing
      ///	- Exactly two same flavour opposite charged leptons
      ///	- Ambiguities in pairing are resolved by choosing the combination
      ///		that results in the smaller value of the sum |mll - mZ| for the two pairs
      /////////////////////////////////////////////////////////////////////////////

      Particles part_pos_el;
      Particles part_neg_el;
      Particles part_pos_mu;
      Particles part_neg_mu;

      foreach( const Particle &l , leptons_sel4l) {
         if(fabs(l.pdgId()) == PID::ELECTRON) {
            if(l.pdgId() < 0) part_neg_el.push_back(l); 
            if(l.pdgId() > 0) part_pos_el.push_back(l); 
         }
         if(fabs(l.pdgId()) == PID::MUON) {
            if(l.pdgId() < 0) part_neg_mu.push_back(l); 
            if(l.pdgId() > 0) part_pos_mu.push_back(l); 
         }
      }

      // ee/mm channel
      if( part_neg_el.size() == 2 || part_neg_mu.size() == 2) {

         Zstate Zcand_1; 
         Zstate Zcand_2; 
         Zstate Zcand_3; 
         Zstate Zcand_4; 

         // ee
         if(part_neg_el.size() == 2) {

            Zcand_1 = Zstate( ParticlePair( part_neg_el[0],  part_pos_el[0] ) ); 
            Zcand_2 = Zstate( ParticlePair( part_neg_el[0],  part_pos_el[1] ) ); 
            Zcand_3 = Zstate( ParticlePair( part_neg_el[1],  part_pos_el[0] ) ); 
            Zcand_4 = Zstate( ParticlePair( part_neg_el[1],  part_pos_el[1] ) ); 

         // mumu
         } else {

            Zcand_1 = Zstate( ParticlePair( part_neg_mu[0],  part_pos_mu[0] ) ); 
            Zcand_2 = Zstate( ParticlePair( part_neg_mu[0],  part_pos_mu[1] ) ); 
            Zcand_3 = Zstate( ParticlePair( part_neg_mu[1],  part_pos_mu[0] ) ); 
            Zcand_4 = Zstate( ParticlePair( part_neg_mu[1],  part_pos_mu[1] ) ); 
         }

         // we can have the following pairs: (Z1 + Z4) || (Z2 + Z3)
         double minValue_1, minValue_2;
         minValue_1 = fabs( Zcand_1.momentum().mass() - Zmass ) + fabs( Zcand_4.momentum().mass() - Zmass );
         minValue_2 = fabs( Zcand_2.momentum().mass() - Zmass ) + fabs( Zcand_3.momentum().mass() - Zmass );   

         if( minValue_1 < minValue_2 ) {
            Z1 = Zcand_1;
            Z2 = Zcand_4;
         } else {
            Z1 = Zcand_2;
            Z2 = Zcand_3;
         }

         // emu channel
      } else if (  part_neg_mu.size() == 1 && part_neg_el.size()==1) {

         Z1 = Zstate ( ParticlePair (part_neg_mu[0],  part_pos_mu[0] ) );
         Z2 = Zstate ( ParticlePair (part_neg_el[0],  part_pos_el[0] ) );

      }

   }

   /// @ZZ
   class ATLAS_2012_I1203852 : public Analysis {
      //protected:

      public:

         /// Default constructor
         ATLAS_2012_I1203852()
            : Analysis("ATLAS_2012_I1203852")
         {
            setNeedsCrossSection(true);

         }

         void init() {

            FinalState fs(-5.0, 5.0, 0.0*GeV);

            vector<pair<double, double> > etaRanges_lep;
            etaRanges_lep.push_back(make_pair(-3.16,3.16));

            // final states to form Z bosons
            vids.push_back(make_pair(PID::ELECTRON, PID::POSITRON));
            vids.push_back(make_pair(PID::MUON, PID::ANTIMUON));

            vnuids.push_back(make_pair(PID::NU_E, PID::NU_EBAR) );
            vnuids.push_back(make_pair(PID::NU_MU, PID::NU_MUBAR) );
            vnuids.push_back(make_pair(PID::NU_TAU, PID::NU_TAUBAR) );

            IdentifiedFinalState Photon(fs);
            Photon.acceptIdPair(PID::PHOTON);

            IdentifiedFinalState bare_EL(fs);
            bare_EL.acceptIdPair(PID::ELECTRON);

            IdentifiedFinalState bare_MU(fs);
            bare_MU.acceptIdPair(PID::MUON);


            // Selection1: ZZ-> llll selection
            DressedLeptons electron_sel4l(Photon, bare_EL, 0.1,true/*Do clustering*/, etaRanges_lep, 7*GeV, false/*false = ignore photons from hadron or tau*/);
            addProjection(electron_sel4l, "ELECTRON_sel4l");

            DressedLeptons muon_sel4l(Photon, bare_MU, 0.1,true, etaRanges_lep, 7*GeV, false);
            addProjection(muon_sel4l, "MUON_sel4l");


            // Selection 2: ZZ-> llnunu selection
            vector<pair<double, double> > etaRanges_lep2;
            etaRanges_lep2.push_back(make_pair(-2.5,2.5));

            DressedLeptons electron_sel2l2nu(Photon, bare_EL, 0.1,true, etaRanges_lep2, 10*GeV, false);
            addProjection(electron_sel2l2nu, "ELECTRON_sel2l2nu");

            DressedLeptons muon_sel2l2nu(Photon, bare_MU, 0.1,true, etaRanges_lep2, 10*GeV, false);
            addProjection(muon_sel2l2nu, "MUON_sel2l2nu");

            /// Get all neutrinos. These will not be used to form jets.
            /// We'll use the highest 2 pT neutrinos to calculate the MET
            IdentifiedFinalState neutrino_fs(-4.5, 4.5, 0.0*GeV);
            neutrino_fs.acceptNeutrinos();
            addProjection(neutrino_fs, "NEUTRINO_FS");

            VetoedFinalState jetinput;
            jetinput.addVetoOnThisFinalState(bare_MU);
            jetinput.addVetoOnThisFinalState(neutrino_fs);

            FastJets jetpro(fs, FastJets::ANTIKT, 0.4);
            addProjection(jetpro, "jet");

            // both ZZ onshel histos         
            _h_ZZ_xsect = bookHisto1D(1, 1, 1);
            _h_ZZ_ZpT   = bookHisto1D(3, 1, 1);
            _h_ZZ_phill = bookHisto1D(5, 1, 1);
            _h_ZZ_mZZ   = bookHisto1D(7, 1, 1);

            // one Z offshell (ZZstar) histos
            _h_ZZs_xsect = bookHisto1D(1, 1, 2);

            // ZZ-> llnunu histos
            _h_ZZnunu_xsect = bookHisto1D(1, 1, 3);
            _h_ZZnunu_ZpT   = bookHisto1D(4, 1, 1);
            _h_ZZnunu_phill = bookHisto1D(6, 1, 1);
            _h_ZZnunu_mZZ   = bookHisto1D(8, 1, 1);

         }

         /// Do the analysis..............................................................................
         void analyze(const Event & e) {
            const double weight = e.weight();

            ////////////////////////////////////////////////////////////////////
            // preselection of leptons for ZZ-> llll final state
            ////////////////////////////////////////////////////////////////////

            Particles leptons_sel4l;  // output

            const vector<ClusteredLepton> & mu_sel4l   = applyProjection<DressedLeptons>(e, "MUON_sel4l").clusteredLeptons();
            const vector<ClusteredLepton> & el_sel4l   = applyProjection<DressedLeptons>(e, "ELECTRON_sel4l").clusteredLeptons();

            vector<ClusteredLepton> leptonsFS_sel4l;
            leptonsFS_sel4l.insert( leptonsFS_sel4l.end(), mu_sel4l.begin(), mu_sel4l.end() );
            leptonsFS_sel4l.insert( leptonsFS_sel4l.end(), el_sel4l.begin(), el_sel4l.end() );

            ////////////////////////////////////////////////////////////////////
            // OVERLAP removal dR(l,l)>0.2
            ////////////////////////////////////////////////////////////////////

            foreach( const ClusteredLepton & l1, leptonsFS_sel4l) {
               bool isolated = true;

               foreach(ClusteredLepton & l2, leptonsFS_sel4l) {

                  const double dR = deltaR(l1, l2);
                  if( (dR < 0.2) && (l1 != l2) ) { isolated = false; break; }

               }

               if(isolated){
                  leptons_sel4l.push_back(l1);
               }
            }

            //////////////////////////////////////////////////////////////////	
            //exactly two opposite charged leptons
            //////////////////////////////////////////////////////////////////

            // calculate total 'flavour' charge
            float totalcharge = 0;
            foreach( Particle &l, leptons_sel4l) { 
               totalcharge += l.pdgId();
            }

            if(leptons_sel4l.size() == 4 && totalcharge == 0  ) {
               // analyze 4 lepton events

               Zstate Z1, Z2;

               // identifies Z states from 4 lepton pairs
               identifyZstates(Z1, Z2,leptons_sel4l);

               ////////////////////////////////////////////////////////////////////////////
               // Z MASS WINDOW
               //	-ZZ: for both Z: 66<mZ<116 GeV
               //	-ZZ*: one Z on-shell: 66<mZ<116 GeV, one Z off-shell: mZ>20 GeV
               ///////////////////////////////////////////////////////////////////////////

               vector<Zstate> zstates_sortpt;
               zstates_sortpt.push_back(Z1);
               zstates_sortpt.push_back(Z2);
               sort(zstates_sortpt.begin()     , zstates_sortpt.end()    , Zstate_less_pT()     );

               Zstate & leadPtZ = zstates_sortpt[0];

               float mZ1   = Z1.momentum().mass();
               float mZ2   = Z2.momentum().mass();
               float ZpT   = leadPtZ.momentum().pT();
               float phill = fabs(deltaPhi(leadPtZ.first.momentum().phi(), leadPtZ.second.momentum().phi() ));
               if(phill > pi) phill = 2*pi-phill;	
               float mZZ   = (Z1.momentum() + Z2.momentum()).mass();               

               if(mZ1 > 20*GeV && mZ2> 20*GeV) {

                  // ZZ* selection
                  if(  (mZ1 > 66*GeV && mZ1 < 116*GeV) || (mZ2 > 66*GeV && mZ2 < 116*GeV ) ){

                     _h_ZZs_xsect  -> fill(sqrtS()*GeV,  weight);

                  }

                  // ZZ selection
                  if(  (mZ1 > 66*GeV && mZ1 < 116*GeV) && (mZ2 > 66*GeV && mZ2 < 116*GeV ) ){

                     _h_ZZ_xsect  -> fill(sqrtS()*GeV,  weight);
                     _h_ZZ_ZpT    -> fill(ZpT   , weight);
                     _h_ZZ_phill  -> fill(phill , weight);
                     _h_ZZ_mZZ    -> fill(mZZ   , weight);

                  }
               }
            }

            ////////////////////////////////////////////////////////////////////
            /// preselection of leptons for ZZ-> llnunu final state
            ////////////////////////////////////////////////////////////////////

            Particles leptons_sel2l2nu;// output

            const vector<ClusteredLepton> & mu_sel2l2nu   = applyProjection<DressedLeptons>(e, "MUON_sel2l2nu").clusteredLeptons();
            const vector<ClusteredLepton> & el_sel2l2nu   = applyProjection<DressedLeptons>(e, "ELECTRON_sel2l2nu").clusteredLeptons();

            vector<ClusteredLepton> leptonsFS_sel2l2nu;


            leptonsFS_sel2l2nu.insert( leptonsFS_sel2l2nu.end(), mu_sel2l2nu.begin(), mu_sel2l2nu.end() );
            leptonsFS_sel2l2nu.insert( leptonsFS_sel2l2nu.end(), el_sel2l2nu.begin(), el_sel2l2nu.end() );

            // lepton preselection for ZZ-> llnunu
            if(  !(mu_sel2l2nu.size() && el_sel2l2nu.size())                                 /* cannot have oposite flavour */
                  && (leptonsFS_sel2l2nu.size() == 2)                                        /* exactly two leptons */
                  && (leptonsFS_sel2l2nu[0].charge() * leptonsFS_sel2l2nu[1].charge() < 1 )  /* opposite charge */
                  && (deltaR(leptonsFS_sel2l2nu[0], leptonsFS_sel2l2nu[1]) > 0.3)            /* overlap removal */
                  && ( (leptonsFS_sel2l2nu[0].pT() > 20*GeV) && (leptonsFS_sel2l2nu[1].pT() > 20*GeV) )   /* trigger requirement */
              ){

               leptons_sel2l2nu.insert(leptons_sel2l2nu.end(), leptonsFS_sel2l2nu.begin(), leptonsFS_sel2l2nu.end());
            }

            if(!leptons_sel2l2nu.size()) vetoEvent;   // no further analysis, fine to veto

            Particles leptons_sel2l2nu_jetveto;  
            foreach( const ClusteredLepton &l, mu_sel2l2nu ) leptons_sel2l2nu_jetveto.push_back(l.constituentLepton());
            foreach( const ClusteredLepton &l, el_sel2l2nu ) leptons_sel2l2nu_jetveto.push_back(l.constituentLepton());


            float ptll = (leptons_sel2l2nu[0].momentum()+leptons_sel2l2nu[1].momentum()).pT();

            // find Z1-> ll 
            FinalState fs2(-3.2, 3.2);
            InvMassFinalState imfs(fs2, vids, 20*GeV, sqrtS()*GeV);
            imfs.calc(leptons_sel2l2nu);

            if(imfs.particlePairs().size() != 1) vetoEvent;

            const ParticlePair & Z1constituents = imfs.particlePairs()[0];
            FourMomentum Z1 = Z1constituents.first.momentum() + Z1constituents.second.momentum();


            // find Z2-> nunu
            const Particles& neutrinoFS = applyProjection<IdentifiedFinalState>(e, "NEUTRINO_FS").particlesByPt();
            FinalState fs3(-4.5, 4.5);
            InvMassFinalState imfs_Znunu(fs3, vnuids, 20*GeV, sqrtS()*GeV, Zmass );
            imfs_Znunu.calc(neutrinoFS);


            if(imfs_Znunu.particlePairs().size() == 0 ) vetoEvent; 

            // Z to neutrinos
            FourMomentum Z2 = imfs_Znunu.particlePairs()[0].first.momentum() + imfs_Znunu.particlePairs()[0].second.momentum();


            float met_Znunu = Z2.pT();

            // mTZZ
            float mT2_1st_term = sqrt(Zmass*Zmass + ptll*ptll) + sqrt(Zmass*Zmass + met_Znunu*met_Znunu);
            float mT2_2nd_term = Z1.px() + Z2.px();
            float mT2_3rd_term = Z1.py() + Z2.py();
            float mT2_1st_term2 = mT2_1st_term*mT2_1st_term;
            float mT2_2nd_term2 = mT2_2nd_term*mT2_2nd_term;
            float mT2_3rd_term2 = mT2_3rd_term*mT2_3rd_term;
            float mTZZ = sqrt(mT2_1st_term2-mT2_2nd_term2-mT2_3rd_term2);

            if(  (Z2.mass() < 66*GeV)  ||  (Z2.mass() > 116*GeV) ) vetoEvent;
            if(  (Z1.mass() < 76*GeV)  ||  (Z1.mass() > 106*GeV) ) vetoEvent;

            /////////////////////////////////////////////////////////////
            // AXIAL MET < 75 GeV
            ////////////////////////////////////////////////////////////

            double dPhiZ1Z2	= fabs(deltaPhi(Z1,Z2));			
            if(dPhiZ1Z2 > (pi)) dPhiZ1Z2 = 2*pi-dPhiZ1Z2;			
            double axialEtmiss = -Z2.pT()*cos(dPhiZ1Z2);

            if( axialEtmiss < 75*GeV ) vetoEvent;

            float ZpT   = Z1.pT();
            float phill = fabs(deltaPhi(Z1constituents.first.momentum().phi(),
                     Z1constituents.second.momentum().phi() ));

            if(phill > (pi)) phill = 2*pi-phill;	

            ////////////////////////////////////////////////////////////////////////////
            // JETS
            //	-"j": found by "jetpro" projection && pT()>25 GeV && |eta|<4.5
            //	-"goodjets": "j"  && dR(electron/muon,jet)>0.3
            //
            //JETVETO: veto all events with at least one goodjet
            ///////////////////////////////////////////////////////////////////////////

            vector<Jet> good_jets;

            foreach(const Jet& j, applyProjection<FastJets>(e, "jet").jetsByPt(25)){//

               if( fabs(j.eta()) > 4.5 ) continue;

               bool isLepton = 0;
               foreach (const Particle& l, leptons_sel2l2nu_jetveto) {
                  const double dR = deltaR(l.momentum(), j.momentum());
                  if (dR < 0.3) { isLepton = true; break; }
               }
               if (!isLepton) good_jets.push_back(j);

            }

            double n_sel_jets = good_jets.size();

            if(n_sel_jets != 0) vetoEvent;

            /////////////////////////////////////////////////////////////
            //Fractional MET and lepton pair difference: "RatioMet"< 0.4
            ////////////////////////////////////////////////////////////

            double RatioMet = fabs(Z2.pT()-Z1.pT())/Z1.pT();

            if( RatioMet  > 0.4 ) vetoEvent;

            //End of ZZllnunu selection, filling histograms

            _h_ZZnunu_xsect  -> fill(sqrtS()*GeV,  weight);
            _h_ZZnunu_ZpT    -> fill(ZpT   , weight);
            _h_ZZnunu_phill  -> fill(phill , weight);
            _h_ZZnunu_mZZ    -> fill(mTZZ   , weight);

         }

         /// Finalize
         void finalize() {

            double norm = crossSection()/sumOfWeights()/femtobarn;

            scale(_h_ZZ_xsect , norm);
            normalize(_h_ZZ_ZpT,  1);
            normalize(_h_ZZ_phill  , 1);
            normalize(_h_ZZ_mZZ    , 1);

            scale(_h_ZZs_xsect , norm);
            scale(_h_ZZnunu_xsect , norm);
            normalize(_h_ZZnunu_ZpT    , 1);
            normalize(_h_ZZnunu_phill  , 1);
            normalize(_h_ZZnunu_mZZ    , 1);

         }

      private:

         Histo1DPtr _h_ZZ_xsect;
         Histo1DPtr _h_ZZ_ZpT;
         Histo1DPtr _h_ZZ_phill;
         Histo1DPtr _h_ZZ_mZZ;

         Histo1DPtr _h_ZZs_xsect;

         Histo1DPtr _h_ZZnunu_xsect;
         Histo1DPtr _h_ZZnunu_ZpT;
         Histo1DPtr _h_ZZnunu_phill;
         Histo1DPtr _h_ZZnunu_mZZ;

         vector<pair<PdgId,PdgId> > vids;
         vector<pair<PdgId,PdgId> > vnuids;

   };


   // The hook for the plugin system
   DECLARE_RIVET_PLUGIN(ATLAS_2012_I1203852);

}
